
/* ====================================================================
   Copyright 2002-2004   Apache Software Foundation

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */
        


package org.apache.poi.hslf.model;

import java.util.*;
import java.io.*;

import org.apache.poi.poifs.filesystem.POIFSFileSystem;
import org.apache.poi.poifs.filesystem.POIFSDocument;
import org.apache.poi.poifs.filesystem.DocumentEntry;

import org.apache.poi.util.LittleEndian;

/**
 * This class contains the main functionality for the Powerpoint file 
 * "reader". It is only a very basic class for now
 *
 * TODO:
 *  - figure out where the sheets and metasheets live from information
 *    in the file, rather than by scanning through the data
 *  - correctly figure out if a text run is unicode, rather than just
 *    guessing based on 0x00's
 *  - correctly figure out what character encoding is used for non unicode
 *    text runs, rather than relying on the system default
 *  - figure out how to match notes to their correct sheet
 * 
 * @author Nick Burch
 */

public class SlideShow
{
  private InputStream istream;
  private POIFSFileSystem filesystem;

  private byte[] _docstream;

  private Sheet[] _sheets;
  private MetaSheet[] _msheets;

  /**
   *  right now this function takes one parameter: a ppt file, and outputs
   *  the text it can find for it
   */
  public static void main(String args[]) throws IOException
  {
	SlideShow foo = new SlideShow(args[0]);

	Sheet[] sheets = foo.getSheets();
	for(int i=0; i<sheets.length; i++) {
		Sheet sheet = sheets[i];
		System.out.println("*Sheet " + sheet.getSheetNumber() + ":");
		TextRun[] runs = sheet.getTextRuns();
		for(int j=0; j<runs.length; j++) {
			TextRun run = runs[j];
			System.out.println("*\tText run " + run.getRunType());
			System.out.println("\n" + run.getText() + "\n");
		}
	}
  }

  /**
   * Constructs a Powerpoint document from fileName. Parses the document 
   * and places all the important stuff into data structures.
   *
   * @param fileName The name of the file to read.
   * @throws IOException if there is a problem while parsing the document.
   */
  public SlideShow(String fileName) throws IOException
  {
  	this(new FileInputStream(fileName));
  }
  
  /**
   * Constructs a Powerpoint document from an input stream. Parses the 
   * document and places all the important stuff into data structures.
   *
   * @param inputStream the source of the data
   * @throws IOException if there is a problem while parsing the document.
   */
  public SlideShow(InputStream inputStream) throws IOException
  {
        //do Ole stuff
		this(new POIFSFileSystem(inputStream));
        istream = inputStream;
  }

  /**
   * Constructs a Powerpoint document from a POIFS Filesystem. Parses the 
   * document and places all the important stuff into data structures.
   *
   * @param filesystem the POIFS FileSystem to read from
   * @throws IOException if there is a problem while parsing the document.
   */
  public SlideShow(POIFSFileSystem filesystem) throws IOException
  {
		this.filesystem = filesystem;

        // Go find a PowerPoint document in the stream
        // Save anything useful we come across
        readFIB();
  }


  /**
   * Shuts things down. Closes underlying streams etc
   *
   * @throws IOException
   */
  public void close() throws IOException
  {
	if(istream != null) {
		istream.close();
	}
	filesystem = null;
  }


  /**
   * Extracts the main document stream from the POI file then hands off 
   * to other functions that parse other areas.
   *
   * @throws IOException
   */
  private void readFIB() throws IOException
  {
	// Get the main document stream
	DocumentEntry docProps =
		(DocumentEntry)filesystem.getRoot().getEntry("PowerPoint Document");

	// Grab the document stream
	_docstream = new byte[docProps.getSize()];
	filesystem.createDocumentInputStream("PowerPoint Document").read(_docstream);

	// For holding the sheets
	Vector sheets = new Vector(10);
	// For holding the meta sheets
	Vector metaSheets = new Vector(10);

	// For when stepping through, so we know what we have
	Boolean isMeta = null;
	Integer lastStartedAt = null;

	// Loop over the file, looking for sheets and meta sheets
	for(int i=0; i < (_docstream.length - 20); i++) {
		// In case we find one
		Boolean foundIsMeta = null;

		// Look for new pages
		// These look like:
		//     00 f3 00 14 00 00 00 ##
		// (where ## is the page number, as a byte)
		if(_docstream[i] == 0x00 &&
			_docstream[i+1] == (0xf3-256) &&
			_docstream[i+2] == 0x03 &&
			_docstream[i+3] == 0x14 &&
			_docstream[i+4] == 0x00 &&
			_docstream[i+5] == 0x00 &&
			_docstream[i+6] == 0x00 
		) {
			// Sheets seem to contain:
			// 28 bytes from i:
			//   00 9f 0f 04 00 00 00 ## 00 00 00 00 00 
			//   a0 0f (len len) 00 00 (text text text)
			// some crap
			//   00 9f 0f 04 00 00 00 ## 00 00 00 00 00
			//   a0 0f (len len) 00 00 (more text text)
			//
			// First ## is 00 (page title), rest can be any number

			// Have yet to spot what defines if unicode text
			//  (xx 00 xx 00) or normal text (xx xx xx)

			//System.out.println("Found new sheet at " + i);
			//System.out.println("\tSheet number is " + _docstream[i+7]);
			foundIsMeta = Boolean.FALSE;
		}

		// Look for meta page entries
		// These look like:
		//     00 c3 0b 08 00 00 00 tt 00 00 00 ww xx yy 0z 0f 00 0d f0 ##
		// tt is type:
		//			00 = master page, notes pages type b
		//			01 = notes page type c or 2nd part of master page 
		//			     or subtitle master page
		//			03 = handout master page
		//	ww xx is one of:
		//			0b 00 = notes page type b
		//			0c 00 = notes page type c
		//			01 00 = master page
		//			02 00 = 2nd part of master page
		//			04 02 = subtitle master page
		//			06 02 = handout page
		// yy is unknown
		// 0z is one of:
		//			00 for 2nd part of master page
		//			01 for notes page, master page, handout page
		//			05 for subtitle master page
		// ## is unknown, but might be a length
		//
		// Other spotted ones include
		//		00   0A 02  01
		//      01   07 00  01
		//      02   09 02  05
		//      03   08 02  05
		//      04   09 02  01
		//      05   08 02  01
		if(_docstream[i] == 0x00 &&
			_docstream[i+1] == (0xc3-256) &&
			_docstream[i+2] == 0x0b &&
			_docstream[i+3] == 0x08 &&
			_docstream[i+4] == 0x00 &&
			_docstream[i+5] == 0x00 &&
			_docstream[i+6] == 0x00 
		) {
			//System.out.println("Found new meta sheet at " + i);
			foundIsMeta = Boolean.TRUE;
		}

		// Create the appropriate sheet if found
		if(foundIsMeta != null) {
			// Is there one to close off and form?
			if(lastStartedAt != null) {
				int start = lastStartedAt.intValue();
				int end = i-1;
				if(isMeta == Boolean.TRUE) {
					metaSheets.add(new MetaSheet(_docstream,start,(end-start)));
				} else {
					sheets.add(new Sheet(_docstream,start,(end-start)));
				}
			}

			lastStartedAt = new Integer(i);
			isMeta = foundIsMeta;
		}
	}

	// Finish off any outstanding sheets
	if(lastStartedAt != null) {
		int start = lastStartedAt.intValue();
		int end = _docstream.length;
		if(isMeta == Boolean.TRUE) {
			metaSheets.add(new MetaSheet(_docstream,start,(end-start)));
		} else {
			sheets.add(new Sheet(_docstream,start,(end-start)));
		}
	}

	// Put the vectors into arrays
	_sheets = new Sheet[sheets.size()];
	_msheets = new MetaSheet[metaSheets.size()];
	for(int i=0; i<_sheets.length; i++) {
		_sheets[i] = (Sheet)sheets.get(i);
	}
	for(int i=0; i<_msheets.length; i++) {
		_msheets[i] = (MetaSheet)metaSheets.get(i);
	}
  }


  // Accesser methods follow

  /**
   * Returns an array of all the normal Sheets found in the slideshow
   */
  public Sheet[] getSheets() { return _sheets; }

  /**
   * Returns an array of all the meta Sheets found in the slideshow
   */
  public MetaSheet[] getMetaSheets() { return _msheets; }
}
