
/* ====================================================================
   Copyright 2002-2004   Apache Software Foundation

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */
        


package org.apache.poi.hslf.model;

import java.util.*;
import java.io.*;

import org.apache.poi.util.LittleEndian;

/**
 * This class represents a run of text in a powerpoint document. That
 *  run could be text on a sheet, or text in a note.
 *  It is only a very basic class for now
 *
 * @author Nick Burch
 */

public class TextRun
{
  public static int UNKNOWN_TEXT = -1;
  public static int SHEET_TEXT = 0;
  public static int NOTE_TEXT = 8;

  private byte[] _run;
  private Boolean _isUnicode = null;
  private String _rawText;
  private String _text;
  private int _textType;
  private byte _runType;

  /**
   * Constructs a Text Run from an underlying sheet or meta sheet.
   * Extracts the text into a string, handling unicode stuff if needed
   *
   * @param bytes the byte array to get data from
   * @param start the point in the byte array to start at
   * @param length how far into the byte array to read
   */
  public TextRun(byte[] bytes, int start, int length)
  {
	_run = new byte[length];
	System.arraycopy(bytes,start,_run,0,length);

	// Zip forward, and see what it is
	// 00 9f 0f 04 00 00 00 ##
	//  00 00 00 00 00 [A0 | A8] 0f (len len) 00 00 text text
	_runType = _run[7];
	if(_run[13] == (0xa0 - 256)) {
		// note
		_textType = NOTE_TEXT;
	} else if(_run[13] == (0xa8 - 256)) {
		// text
		_textType = SHEET_TEXT;
	} else {
		// who knows
		System.err.println("Hit unknown text text, run type is " + _runType + ", text type is " + _run[13]);
		_textType = UNKNOWN_TEXT;
	}

	// Figure out how long the text is
	long size = LittleEndian.getUInt(_run,15);

	// Is it unicode?
	// ** WARNING **
	//   this only works if the first two characters don't have a high-byte
	//   component. Will fail otherwise! Really need to find the unicode
	//   status from somewhere in the document
	if(_isUnicode == null) {
		if(_run[20] == 0 && _run[22] == 0) {
			_isUnicode = Boolean.TRUE;
		} else {
			_isUnicode = Boolean.FALSE;
		}
	}

	// Create the text from the bytes
	if(_isUnicode == Boolean.TRUE) {
		try {
			//System.out.println("Found unicode text");
			_rawText = pptUnicodeMakeString(_run,19,(int)size);
		} catch(UnsupportedEncodingException uee) {
			throw new Error("Hit unicode text, but utf-16 support not present:" + uee);
		}
	} else {
		_rawText = new String(_run,19,(int)size);
	}

	// PowerPoint seems to store files with \r as the line break
	// The messes things up on everything but a Mac, so translate
	//  them to \n
	_text = _rawText.replace('\r','\n');
  }


  /**
   * Constructs a String from a unicode based length of bytes. Java
   *  assumes the format for UTF-16 is (high)(low), but PPT stores it
   *  as (low)(high), yey...
   *
   * @param b the byte array to get data from
   * @param start the point in the byte array to start at
   * @param len how far into the byte array to read
   */
  public static String pptUnicodeMakeString(byte[] b, int start, int len) throws UnsupportedEncodingException {
	byte[] reversed = new byte[len];
	for(int i=0; i<len; i+= 2) {
		reversed[i] = b[i+start+1];
		reversed[i+1] = b[i+start];
	}
	return new String(reversed,"UTF-16");
  }


  // Accesser methods follow

  /**
   * Returns the text content of the run
   */
  public String getText() { return _text; }

  /**
   * Returns the raw text content of the run. This hasn't had any
   *  changes applied to it, and so is probably unlikely to print
   *  out nicely.
   */
  public String getRawText() { return _rawText; }

  /**
   * Returns the type of the run, be it SHEET_TEXT or NOTE_TEXT
   */
  public int getTextType() { return _textType; }

  /**
   * Returns the raw type of the run, from offset 7
   */
  public byte getRunType() { return _runType; }

} 
